package com.nisum.reactiveweb.controller;

import com.nisum.reactiveweb.model.Product;
import com.nisum.reactiveweb.service.ProductService;
import lombok.extern.slf4j.Slf4j;

import java.util.stream.Collectors;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.support.DefaultMessageSourceResolvable;
import org.springframework.util.Assert;
import org.springframework.validation.BeanPropertyBindingResult;
import org.springframework.validation.Errors;
import org.springframework.validation.Validator;
import org.springframework.web.reactive.function.server.RouterFunction;
import org.springframework.web.reactive.function.server.ServerRequest;
import org.springframework.web.reactive.function.server.ServerResponse;

import static org.springframework.web.reactive.function.server.RouterFunctions.route;
import static org.springframework.web.reactive.function.server.ServerResponse.accepted;
import static org.springframework.web.reactive.function.server.ServerResponse.created;
import static org.springframework.web.reactive.function.server.ServerResponse.ok;

@Slf4j
@Configuration
public class ProductHandler {

    @Autowired
    private Validator validator;

    private static final String SKU_TEXT = "sku";

    private static final String SKU_TEXT_PATH_VAR = "/{sku}";

    @Bean
    public RouterFunction<ServerResponse> routes(ProductService productService) {
        return route()
                .before(this::logAndReturn)
                .path("/product", () -> route()
                        .GET(SKU_TEXT_PATH_VAR,
                             req -> productService.getProductBySku(req.pathVariable(SKU_TEXT))
                                     .flatMap(product -> ok()
                                             .bodyValue(product))
                        )
                        .POST("",
                              req -> req.bodyToMono(Product.class)
                                      .doOnNext(this::validateProduct)
                                      .flatMap(productService::saveProduct)
                                      .flatMap(product -> created(req.uri())
                                              .bodyValue(product))
                        )
                        .PUT(SKU_TEXT_PATH_VAR,
                             req -> req.bodyToMono(Product.class)
                                     .doOnNext(this::validateProduct)
                                     .flatMap(product -> productService.updateProduct(req.pathVariable(SKU_TEXT), product))
                                     .flatMap(product -> accepted()
                                             .bodyValue(product))
                        )
                        .DELETE(SKU_TEXT_PATH_VAR,
                                req -> productService.removeProduct(req.pathVariable(SKU_TEXT))
                                        .flatMap(res -> ok().bodyValue(res))
                        )
                        .build())
                .path("/products", () -> route()
                        .GET("",
                             req -> ok()
                                     .body(productService.getAllProducts(), Product.class)
                        ).build())
                .build();
    }

    private ServerRequest logAndReturn(ServerRequest req) {
        log.info("Request : {}", req.toString());
        return req;
    }

    private void validateProduct(Product product) {
        Assert.notNull(product, "product can't be null");
        Errors errors = new BeanPropertyBindingResult(product, "product");
        validator.validate(product, errors);
        String errorString = errors.getAllErrors().
                stream().map(DefaultMessageSourceResolvable::getDefaultMessage)
                .collect(Collectors.joining(", "));
        if (errors.hasErrors()) {
            throw new IllegalArgumentException(errorString);
        }
    }

}
