package com.nisum.reactiveweb.controller;

import com.nisum.reactiveweb.exceptions.ApiErrorAttributes;
import com.nisum.reactiveweb.exceptions.ProductNotFoundException;
import com.nisum.reactiveweb.model.Product;
import com.nisum.reactiveweb.service.ProductServiceImpl;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.math.BigDecimal;
import java.util.function.Predicate;

import org.hamcrest.Matchers;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.reactive.WebFluxTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.context.annotation.Import;
import org.springframework.test.web.reactive.server.WebTestClient;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.Mockito.when;
import static reactor.core.publisher.Mono.just;

@WebFluxTest()
@Import({ProductHandler.class, ApiErrorAttributes.class})
class ProductHandlerComponentTests {

    @Autowired
    private WebTestClient testClient;

    @MockBean
    private ProductServiceImpl productService;

    private final Product product1 = new Product("1", "toy", new BigDecimal(200));

    private final Product product3 = new Product("1", "toy", new BigDecimal(300));

    private final Product product2 = new Product("2", "table", new BigDecimal(400));

    private final Predicate<Product> productPredicate1 = productPredicate("1", "toy", 200);

    private final Predicate<Product> productPredicate2 = productPredicate("1", "toy", 300);

    private final Mono<Product> productNotFoundErrorMono = Mono.error(ProductNotFoundException::new);

    @Test
    void getProductBySku() {
        when(productService.getProductBySku("1")).thenReturn(productMono());
        testClient.get()
                .uri("/product/1")
                .exchange()
                .expectStatus().isOk()
                .expectBody(Product.class).value(Matchers.equalTo(product1));
    }

    @Test
    void getProductBySkuWhenProductDoesNotExist() {
        when(productService.getProductBySku("1")).thenReturn(productNotFoundErrorMono);
        testClient.get()
                .uri("/product/1")
                .exchange()
                .expectStatus().isNotFound();
    }

    @Test
    void saveProduct() {
        Product productToSave = new Product(null, "toy", new BigDecimal(200));
        when(productService.saveProduct(productToSave)).thenReturn(productMono());
        testClient.post()
                .uri("/product")
                .body(just(productToSave), Product.class)
                .exchange()
                .expectStatus().isCreated()
                .expectBody(Product.class).isEqualTo(product1)
                .consumeWith(result -> {
                    Product savedProduct = result.getResponseBody();
                    assertNotNull(savedProduct);
                    assertNotNull(savedProduct.getSku());
                    assertTrue(productPredicate1.test(savedProduct));
                });
    }

    @Test
    void saveProductWithProductNameOrPriceAsNull() {
        Product productToSave = new Product();
        testClient.post()
                .uri("/product")
                .body(just(productToSave), Product.class)
                .exchange()
                .expectStatus().isBadRequest();
    }

    @Test
    void updateProduct() {
        Product productToUpdate = new Product(null, "toy", new BigDecimal(300));
        String sku = "1";
        when(productService.updateProduct(sku, productToUpdate)).thenReturn(Mono.just(product3));
        testClient.put()
                .uri("/product/" + sku)
                .body(just(productToUpdate), Product.class)
                .exchange()
                .expectStatus().isAccepted()
                .expectBody(Product.class).isEqualTo(product3)
                .consumeWith(result -> {
                    Product updatedProduct = result.getResponseBody();
                    assertNotNull(updatedProduct);
                    assertEquals(sku, updatedProduct.getSku());
                    assertEquals(product3.getPrice(), updatedProduct.getPrice());
                    assertTrue(productPredicate2.test(updatedProduct));
                });
    }

    @Test
    void updateProductWhenProductNameOrPriceIsNull() {
        Product productToUpdate = new Product();
        String sku = "5";
        testClient.put()
                .uri("/product/" + sku)
                .body(just(productToUpdate), Product.class)
                .exchange()
                .expectStatus().isBadRequest();
    }

    @Test
    void updateProductWhenProductDoesNotExistWithGivenSku() {
        Product productToUpdate = new Product(null, "toy", new BigDecimal(300));
        String sku = "5";
        when(productService.updateProduct(sku, productToUpdate)).thenReturn(productNotFoundErrorMono);
        testClient.put()
                .uri("/product/" + sku)
                .body(just(productToUpdate), Product.class)
                .exchange()
                .expectStatus().isNotFound();
    }

    @Test
    void removeProduct() {
        String sku = "5";
        final String deletedText = "deleted";
        when(productService.removeProduct(sku)).thenReturn(Mono.just(deletedText));
        testClient.delete()
                .uri("/product/" + sku)
                .exchange()
                .expectStatus().isOk()
                .expectBody(String.class)
                .consumeWith(result -> assertEquals(deletedText, result.getResponseBody()));
    }

    @Test
    void removeProductWhenSkuDoesNotExist() {
        String sku = "5";
        when(productService.removeProduct(sku)).thenReturn(Mono.error(ProductNotFoundException::new));
        testClient.delete()
                .uri("/product/" + sku)
                .exchange()
                .expectStatus().isNotFound();
    }

    @Test
    void getAllProducts() {
        when(productService.getAllProducts()).thenReturn(productsFlux());

        testClient.get().uri("/products")
                .exchange()
                .expectStatus().isOk()
                .expectBodyList(Product.class)
                .hasSize(2)
                .contains(product1, product2);

    }

    private Mono<Product> productMono() {
        return just(product1);
    }

    private Flux<Product> productsFlux() {
        return Flux.just(product1, product2);
    }

    private Predicate<Product> productPredicate(String sku, String name, int price) {
        return product -> product.getSku().equals(sku)
                && product.getName().equals(name)
                && product.getPrice().intValue() == price;
    }

}