package com.nisum.reactiveweb.controller;

import com.nisum.reactiveweb.model.Product;
import com.nisum.reactiveweb.repository.ProductRepository;

import java.math.BigDecimal;

import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.MethodOrderer.OrderAnnotation;
import org.junit.jupiter.api.Order;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.TestMethodOrder;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.reactive.AutoConfigureWebTestClient;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.web.reactive.server.WebTestClient;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static reactor.core.publisher.Mono.just;

@SpringBootTest
@ActiveProfiles("test")
@TestMethodOrder(OrderAnnotation.class)
@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@AutoConfigureWebTestClient
class ProductHandlerIntegrationTests {

    @Autowired
    private WebTestClient testClient;

    @Autowired
    private ProductRepository repository;

    private String sku = "";

    @BeforeAll
    @AfterAll
    void setup() {
        repository.removeAllProducts().block();
    }

    @Test
    @Order(3)
    void getProductBySku() {
        testClient.get()
                .uri("/product/" + sku)
                .exchange()
                .expectStatus().isOk()
                .expectBody(Product.class)
                .consumeWith(result -> {
                    Product product = result.getResponseBody();
                    assertNotNull(product);
                });
    }

    @Test
    void getProductBySkuWhenProductDoesNotExist() {
        testClient.get()
                .uri("/product/1")
                .exchange()
                .expectStatus().isNotFound();
    }

    @Test
    @Order(1)
    void saveProduct() {
        Product productToSave = new Product(null, "toy", new BigDecimal(200));
        testClient.post()
                .uri("/product")
                .body(just(productToSave), Product.class)
                .exchange()
                .expectStatus().isCreated()
                .expectBody(Product.class)
                .consumeWith(result -> {
                    Product savedProduct = result.getResponseBody();
                    assertNotNull(savedProduct);
                    assertNotNull(savedProduct.getSku());
                    sku = savedProduct.getSku();
                    assertEquals(productToSave.getName(), savedProduct.getName());
                    assertEquals(productToSave.getPrice(), savedProduct.getPrice());
                });
    }

    @Test
    void saveProductWithProductNameOrPriceAsNull() {
        Product productToSave = new Product();
        testClient.post()
                .uri("/product")
                .body(just(productToSave), Product.class)
                .exchange()
                .expectStatus().isBadRequest();
    }

    @Test
    @Order(2)
    void updateProduct() {
        String productName = "toy";
        BigDecimal productPrice = new BigDecimal(300);
        Product productToUpdate = new Product(null, productName, productPrice);
        testClient.put()
                .uri("/product/" + sku)
                .body(just(productToUpdate), Product.class)
                .exchange()
                .expectStatus().isAccepted()
                .expectBody(Product.class)
                .consumeWith(result -> {
                    Product updatedProduct = result.getResponseBody();
                    assertNotNull(updatedProduct);
                    assertEquals(sku, updatedProduct.getSku());
                    assertEquals(productName, updatedProduct.getName());
                    assertEquals(productPrice, updatedProduct.getPrice());
                });
    }

    @Test
    void updateProductWhenProductNameOrPriceIsNull() {
        Product productToUpdate = new Product();
        testClient.put()
                .uri("/product/" + sku)
                .body(just(productToUpdate), Product.class)
                .exchange()
                .expectStatus().isBadRequest();
    }

    @Test
    void updateProductWhenProductDoesNotExistWithGivenSku() {
        Product productToUpdate = new Product(null, "toy", new BigDecimal(300));
        String sku = "5";
        testClient.put()
                .uri("/product/" + sku)
                .body(just(productToUpdate), Product.class)
                .exchange()
                .expectStatus().isNotFound();
    }

    @Test
    @Order(10)
    void removeProduct() {
        final String deletedText = "deleted";
        testClient.delete()
                .uri("/product/" + sku)
                .exchange()
                .expectStatus().isOk()
                .expectBody(String.class)
                .consumeWith(result -> assertEquals(deletedText, result.getResponseBody()));
    }

    @Test
    void removeProductWhenSkuDoesNotExist() {
        String sku = "5";
        testClient.delete()
                .uri("/product/" + sku)
                .exchange()
                .expectStatus().isNotFound();
    }

    @Test
    @Order(3)
    void getAllProducts() {
        testClient.get().uri("/products")
                .exchange()
                .expectStatus().isOk()
                .expectBodyList(Product.class)
                .hasSize(1);
    }

}