package com.nisum.reactiveweb.service;

import com.nisum.reactiveweb.exceptions.ProductNotFoundException;
import com.nisum.reactiveweb.model.Product;
import com.nisum.reactiveweb.repository.ProductRepository;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;
import reactor.test.StepVerifier;

import java.math.BigDecimal;
import java.util.function.Predicate;

import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.boot.test.context.SpringBootTest;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.when;
import static reactor.core.publisher.Mono.just;

@SpringBootTest
class ProductServiceTest {

    @InjectMocks
    private ProductServiceImpl productService;

    @Mock
    private ProductRepository repository;

    private final Product product1 = new Product("1", "toy", new BigDecimal(200));

    private final Product product3 = new Product("1", "toy", new BigDecimal(300));

    private final Product product2 = new Product("2", "table", new BigDecimal(400));

    private final Predicate<Product> productPredicate1 = productPredicate("1", "toy", 200);

    private final Predicate<Product> productPredicate3 = productPredicate("1", "toy", 300);

    private final Predicate<Product> productPredicate2 = productPredicate("2", "table", 400);

    @Test
    void getProductBySkuWhenProductExists() {
        when(repository.getProductBySku("1")).thenReturn(productMono());

        StepVerifier.create(productService.getProductBySku("1"))
                .expectNextMatches(productPredicate1)
                .verifyComplete();
    }

    @Test
    void getProductBySkuWhenProductDoesNotExist() {
        when(repository.getProductBySku("1")).thenReturn(Mono.empty());

        StepVerifier.create(productService.getProductBySku("1"))
                .expectError(ProductNotFoundException.class)
                .verify();
    }

    @Test
    void getAllProducts() {
        when(repository.getAllProducts()).thenReturn(productsFlux());

        StepVerifier.create(productService.getAllProducts())
                .expectNextMatches(productPredicate1)
                .as("predicate1")
                .expectNextMatches(productPredicate2)
                .as("predicate2")
                .verifyComplete();
    }

    @Test
    void saveProduct() {
        when(repository.saveProduct(any(Product.class))).thenReturn(productMono());

        StepVerifier.create(productService.saveProduct(product1))
                .expectNextMatches(productPredicate1)
                .verifyComplete();
    }

    @Test
    void updateProductWhenProductExists() {
        when(repository.getProductBySku(anyString())).thenReturn(productMono());
        when(repository.updateProduct(anyString(), any(Product.class))).thenReturn(just(product3));

        StepVerifier.create(productService.updateProduct("1", product3))
                .expectNextMatches(productPredicate3)
                .verifyComplete();
    }

    @Test
    void updateProductWhenProductDoesNotExist() {
        when(repository.getProductBySku("1")).thenReturn(Mono.empty());
        when(repository.updateProduct(anyString(), any(Product.class))).thenReturn(just(product3));

        StepVerifier.create(productService.updateProduct("1", product3))
                .expectError(ProductNotFoundException.class)
                .verify();
    }

    @Test
    void removeProductWhenProductExists() {
        when(repository.removeProduct(anyString())).thenReturn(just(1L));

        StepVerifier.create(productService.removeProduct("1"))
                .expectNext("deleted")
                .verifyComplete();
    }

    @Test
    void removeProductWhenProductDoesNotExist() {
        when(repository.removeProduct(anyString())).thenReturn(just(0L));

        StepVerifier.create(productService.removeProduct("1"))
                .expectError(ProductNotFoundException.class)
                .verify();
    }

    private Mono<Product> productMono() {
        return just(product1);
    }

    private Flux<Product> productsFlux() {
        return Flux.just(product1, product2);
    }

    private Predicate<Product> productPredicate(String sku, String name, int price) {
        return product -> product.getSku().equals(sku)
                && product.getName().equals(name)
                && product.getPrice().intValue() == price;
    }
}